//MUPS - Septembar 2008 - zadatak 9
//racunanje srednje vrednosti i sd za ucitani niz

#include "mpi.h"
#include <iostream>
#include <cmath>

using namespace std;

enum proc {MASTER, SLAVE};

int main (int argc, char* argv[]) {
	
	//init
	MPI::Init(argc,argv);
	
	//dovtanje ranga i velicine komunikatora
	int rank = MPI::COMM_WORLD.Get_rank();
	int size = MPI::COMM_WORLD.Get_size();

	//ispisi poruke
	cout << "Hello MPI world!" << endl;	
	cout << "Moj rang je: " << rank << endl;
	cout << "Ukupan broj procesa u ovom komunikatoru je: " << size << endl;
	
	//ako si master ucitaj niz, posalji posao, ispisi rezultate
	if (rank == MASTER) {
		
		int n;
		cout << "Unesite duzinu niza: ";
		cin >> n;

		//posalji svima duzinu niza
		MPI::COMM_WORLD.Bcast(&n,1,MPI::INT,MASTER);

		if (n % size != 0) {
			cout << "Duzina niza nije celobrojni umnozak broja procesa!\nPREKIDAM!";		
			MPI::COMM_WORLD.Abort(rank);
		}

		//ucitaj niz;
		int* send_buff = new int[n];
		cout << "\nUnesite elemente niza: " << endl; 
		for(int i=0;i<n;i++)
			cin >> send_buff[i];

		int chunk = n / size;
		int* recv_buff = new int[chunk];

		//posalji po parce niza svima
		MPI::COMM_WORLD.Scatter(send_buff,chunk,MPI::INT,recv_buff,chunk,MPI::INT,MASTER);

		int part_sum=0;
		//sracunaj svoje parce sume
		for(int i=0;i<chunk;i++)
			part_sum+=recv_buff[i];

		int sum;
		//skupi sve parcijalne sume
		MPI::COMM_WORLD.Allreduce(&part_sum,&sum,1,MPI::INT,MPI::SUM);
		
		cout << "Suma: " << sum << endl;

		double avg = (double)sum/n;

		//racunaj svoj deo devijacije
		double part_sd = 0;
		for(int i=0;i<chunk;i++)
			part_sd+=(recv_buff[i]-avg)*(recv_buff[i]-avg);

		//redukujemo rezultate
		double sd;
		MPI::COMM_WORLD.Reduce(&part_sd,&sd,1,MPI::DOUBLE,MPI::SUM,MASTER);
		
		sd = sqrt(sd/n);

		cout << "\nSrednja vrednost niza: " << avg << endl;
		cout << "Standardna devijacija niza: " << sd << endl;
	
	} 
	//ako si slave, cuti i radi
	else {
		int n;

		//primi duzinu niza
		MPI::COMM_WORLD.Bcast(&n,1,MPI::INT,MASTER);
		
		
		if (n % size != 0) {
			cout << "Duzina niza nije celobrojni umnozak broja procesa!\nPREKIDAM!";		
			MPI::COMM_WORLD.Abort(rank);
		}

		int chunk = n / size;
		int* recv_buff = new int[chunk];

		//primi parce niza na obradu
		MPI::COMM_WORLD.Scatter(0,chunk,MPI::INT,recv_buff,chunk,MPI::INT,MASTER);

		int part_sum=0;
		//sracunaj svoje parce sume
		for(int i=0;i<chunk;i++)
			part_sum+=recv_buff[i];

		int sum;
		//skupi sve parcijalne sume
		MPI::COMM_WORLD.Allreduce(&part_sum,&sum,1,MPI::INT,MPI::SUM);

		//cout << "\nParcijalna suma: " << part_sum;
		//cout << "\nSuma: " << sum;
		
		double avg = (double)sum/n;

		//racunaj svoj deo devijacije
		double part_sd = 0;
		for(int i=0;i<chunk;i++)
			part_sd+=(recv_buff[i]-avg)*(recv_buff[i]-avg);

		//redukujemo rezultate
		MPI::COMM_WORLD.Reduce(&part_sd,0,1,MPI::DOUBLE,MPI::SUM,MASTER);
		
	}

	MPI::Finalize();
	return 0;
}