/// @file groups_mpi.c 
/// @brief Primer rada sa MPI grupama.
///
/// Program nema drugi smisao izuzev pokaznog.
/// Vredi obratiti paÅ¾nju na to kako su upareni pozivi slanja i prijema.
/// @warning Program neće imati smisla za bilo koji broj procesa različit od NPROCS.
/// Promena vrednosti NPROCS treba da prouzrokuje i promenu vrednosti elemenata nizova ranks1 i ranks2.

#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>

#define NPROCS 8

int main(int argc, char *argv[])
{
	int rank, new_rank, sendbuf, recvbuf, numtasks;
	int ranks1[4]={0,1,2,3}, ranks2[4]={4,5,6,7};
	MPI_Group  orig_group, new_group;
	MPI_Comm   new_comm;

	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &numtasks);

	printf("Hi there! I am %d of %d\n", rank, numtasks);

	if (numtasks != NPROCS) 
	{
		printf("Must specify MP_PROCS= %d. Terminating.\n",NPROCS);
		MPI_Finalize();
		exit(0);
	}

	sendbuf = rank;

	/* Extract the original group handle */
	MPI_Comm_group(MPI_COMM_WORLD, &orig_group);

	/* Divide tasks into two distinct groups based upon rank */
	if (rank < NPROCS/2) 
	{
		MPI_Group_incl(orig_group, NPROCS/2, ranks1, &new_group);
	}
	else 
	{
		MPI_Group_excl(orig_group, NPROCS/2, ranks1, &new_group);
	}

	/* Create new new communicator and then perform collective communications */
	MPI_Comm_create(MPI_COMM_WORLD, new_group, &new_comm);
	MPI_Allreduce(&sendbuf, &recvbuf, 1, MPI_INT, MPI_SUM, new_comm);

	MPI_Group_rank (new_group, &new_rank);
	printf("oldrank= %d newrank= %d recvbuf= %d\n",rank,new_rank,recvbuf);

	MPI_Finalize();
}
