#include "mpi.h"
#include <stdio.h>

#define ROW_COUNT 2
#define COLUMN_COUNT 3
#define SIZE (ROW_COUNT*COLUMN_COUNT)
#define DIMENSION_COUNT 2

#define NEIGHBOUR_COUNT 4

#define TRUE 1
#define FALSE 0

enum Neighbours 
{
	UP, //    0
	DOWN, //  1
	LEFT, //  2
	RIGHT //  3
};

enum AxisNames
{
	X, //     0
	Y //      1
};

enum ShiftOptions
{
	UPWARD_SHIFT_ONE = 1, // >0
	DOWNWARD_SHIFT_ONE = -UPWARD_SHIFT_ONE // <0
};

int main(int argc, char *argv[])
{
	int num_tasks, rank, cart_rank, source, dest, out_buf, i, tag=1, 
		in_buf[NEIGHBOUR_COUNT] = { 0, 0, 0, 0 }, 
		neighbours[NEIGHBOUR_COUNT] = { MPI_PROC_NULL, MPI_PROC_NULL, MPI_PROC_NULL, MPI_PROC_NULL },
		dimensions[DIMENSION_COUNT] = { ROW_COUNT, COLUMN_COUNT }, 
		dimenions_periods[DIMENSION_COUNT] = { TRUE, FALSE }, coordinates[DIMENSION_COUNT];

	MPI_Request send_requests[NEIGHBOUR_COUNT], receive_requests[NEIGHBOUR_COUNT];
	MPI_Status send_statuses[NEIGHBOUR_COUNT], receive_statuses[NEIGHBOUR_COUNT];
	MPI_Comm cart_comm;

	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &num_tasks);

	if (SIZE == num_tasks) 
	{
		/* napravi odgovarajući Cart komunikator */
		MPI_Cart_create(MPI_COMM_WORLD, DIMENSION_COUNT, dimensions, dimenions_periods, FALSE, &cart_comm);
		/* dohvati rang unutar Cart komunikatora */
		MPI_Comm_rank(cart_comm, &cart_rank);
		/* izračunaj svoje koordinate u matrici/valjku/torusu */
		MPI_Cart_coords(cart_comm, cart_rank, DIMENSION_COUNT, coordinates);
		/* definiši orijentaciju, šta je gore/dole, šta je levo/desno */
		/* pažnja: UP (gore) u matrici je pomeraj prema manjoj vrednosti X koordinate u odnosu na dati element */
		MPI_Cart_shift(cart_comm, X, UPWARD_SHIFT_ONE, &neighbours[UP], &neighbours[DOWN]);
		MPI_Cart_shift(cart_comm, Y, UPWARD_SHIFT_ONE, &neighbours[LEFT], &neighbours[RIGHT]);

		/* spakuj za slanje svoje koordinate kao matematičar: [0][0] -> 11 */
		out_buf = (coordinates[X]+1)*10 + (coordinates[Y]+1);

		for (i=0; i<NEIGHBOUR_COUNT; i++) 
		{
			/* svakom susedu asinhrono pošalji svoj podatak */
			dest = neighbours[i];
			MPI_Isend(&out_buf, 1, MPI_INT, dest, tag, cart_comm, &send_requests[i]);
			/* od svakog suseda asinrono primi njegov podatak */
			source = neighbours[i];
			MPI_Irecv(&in_buf[i], 1, MPI_INT, source, tag, cart_comm, &receive_requests[i]);
		}

		/* sačekaj da sva slanja budu završena */
		MPI_Waitall(NEIGHBOUR_COUNT, send_requests, send_statuses);
		/* sačekaj da svi prijemi budu završeni */
		MPI_Waitall(NEIGHBOUR_COUNT, receive_requests, receive_statuses);

		/* ispiši svoje rangove, svoje koordinate i rangove suseda */
		printf("rank= %d cart_rank %d coordinates=[%d,%d] neighbours(u,d,l,r)= %3d%3d%3d%3d\n",
			rank, cart_rank, coordinates[X], coordinates[Y],
			neighbours[UP], neighbours[DOWN], neighbours[LEFT], neighbours[RIGHT]);
		/* ispiši svoj rank u Cart komunikatoru i primljene podatke od suseda */
		printf("cart_rank= %d                              in_buf(u,d,l,r)= %3d%3d%3d%3d\n\n",
			cart_rank, in_buf[UP], in_buf[DOWN], in_buf[LEFT], in_buf[RIGHT]);
	}
	else
		printf("%d: Must specify %d processors. Terminating.\n", rank, SIZE);

	MPI_Comm_free(&cart_comm);
	MPI_Finalize();
}
