
/******************************************************************************/
/**
@file "scalar_product_openmp.c"
@brief Parallel bucket sort primer.

Racunanje skalarnog proizvoda dva vektora u paraleli, uz pomoc OpenMP
*/
/******************************************************************************/
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>

#define VECT_LEN  10000
/******************************************************************************/

double *a, *b; /* Pokazivaci na nizove koje cemo skalarno pomnoziti */
int	len;   /* Duzina segmenata koje ce svaka nit dobiti na obradu */
double	sum;   /* Promenljiva koja ce sadrzati rezultat */

/******************************************************************************/

void *dot_prod();

/******************************************************************************/


/**
@brief Glavni program
@param argc Broj argumenata komandne linije
@param argv Argumenti komandne linije
@return Vraca @b 0 ako je sve uspesno izvrseno
*/
int main (int argc, char* argv[])
{
	int i, numthreads; 

	/* Zabranjujemo okruzenju da bira broj niti */
	omp_set_dynamic(0); 
	
	/* Neka se svaki put stvara onoliko niti koliko ima procesora... */
	numthreads = omp_get_num_procs();
	omp_set_num_threads(numthreads); 

	/* Inicijalizacija */
	len = VECT_LEN;
	sum=0;
 
	a = (double*) malloc (numthreads * len * sizeof(double));
	b = (double*) malloc (numthreads * len * sizeof(double));
	
	/* Vrednosti nizova mozemo da inicijalizujemo u paraleli ako su nizovi veci od 1000 */
	#pragma omp parallel for schedule(static) shared(a,b) if(len*numthreads>1000)
	for (i = 0; i < len * numthreads; i++){
		a[i] = 1;
		b[i] = 1;
	}
	
	/* Paralelno izvrsavanje funkcije dot_prod(), uz redukciju nad promenljivom sum */
	#pragma omp parallel shared(a,b,len) reduction(+:sum)
	{
	    dot_prod();
	}
	
	printf("Done. Sum  =  %f \n", sum);

	free (a);
	free (b);
	
	return 0;
}

/**
@brief Vrsi mnozenje koordinata
*/
void* dot_prod(){
		
	int i, start, end, myid;

	myid = omp_get_thread_num();
	start = myid * len;
	end   = start + len;
	
	/* Na osnovu rednog broja niti, biramo segmenat koordinata koje cemo mnoziti.
	   Promenljiva sum ce ustvari ovde biti private, i u nju ce biti smesten parcijalni
	   rezultat. Nakon izlaska iz paralelnog regiona, vrednosti parcijalnih rezultata
	   ce biti sabrani u istoimenu deljenu promenljivu.
	 */
	for (i=start; i<end ; i++) 
		sum += (a[i] * b[i]);
	

}

