//------------------------------------------------------------------------------
/**
@file "Pthreads/C++/Primer 1 - thread/SemaphoreCpp.h"
@brief Definicija klase semafora.
*/
//------------------------------------------------------------------------------

#ifndef _Semaphore_h_
#define _Semaphore_h_

//------------------------------------------------------------------------------

#include <pthread.h>

//------------------------------------------------------------------------------

/**
@brief Klasa semafora.
*/
class SemaphoreCpp
{
	public:
		/**
		@brief Konstruktor klase.
		@param initial Inicijalna vrednost semafora.

		Alocira novi semafor i inicijalizuje njegovu vrednost, mutex i 
		uslovnu promenljivu.
		*/
		SemaphoreCpp(int initial)
		{ 
			val = initial; 
			pthread_mutex_init(&mutex, NULL);
			pthread_cond_init(&condition, NULL);
		}

		/**
		@brief Destruktor klase.
	
		Unitava mutex i uslovnu promenljivu.
		*/
		~SemaphoreCpp()
		{
			pthread_mutex_destroy(&mutex);
			pthread_cond_destroy(&condition);
		}

		/**
		@brief Implementacija wait metode.

		Nalaze niti da pusti pusti bravu objekta kada je vrednost semafora 0.
		*/
		void wait()
		{
			pthread_mutex_lock(&mutex);

			while (val == 0) pthread_cond_wait(&condition, &mutex);
		
			--val;

			pthread_mutex_unlock(&mutex);
		}

		/**
		@brief Implementacija signal metode.

		Inkrementira vrednost semafora. Ako je vrednost semafora jednaka 1
		odblokira makar jednu nit koja ceka blokirana da se neki uslov ispuni.
		*/
		void signal()
		{
			pthread_mutex_lock(&mutex);
			++val;
			if (val == 1) pthread_cond_signal(&condition);
		
			pthread_mutex_unlock(&mutex);
		}
		
	protected:
		/**
		@property val
		@brief Vrednost semafora.
		*/
		int val;

		/**
		@property mutex
		@brief Mutex promenljiva.
		*/
		pthread_mutex_t mutex;

		/**
		@property condition
		@brief Uslovna promenljiva.
		*/
		pthread_cond_t condition;
};

//------------------------------------------------------------------------------

#endif

//------------------------------------------------------------------------------
