/******************************************************************************/
/**
@file "Pthreads/C/Ispitni zadaci/mps_sep_2008_z7.c"
@brief Prebrojavanje novca po dzakovima.

@author Marko Mii
*/
/******************************************************************************/

//MUPS - Septembar 2008 - zadatak 7
//prebrojavanje novca po dzakovima

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

//promenljive u kojima su smesteni podaci o dzakovima
int** bags;
//niz "duzina" dzakova
int* bagsCount;
//pojedinacne sume za svaki dzak
int* bagsSum;

//brava i deljena promenljiva
pthread_mutex_t moneyMutex;
int moneySum;

void* countTheMoney (void* tId) {
	
	int i;
	int myBag = (int)tId;

	bagsSum[myBag] = 0;

	for(i=0;i<bagsCount[myBag];i++)
		bagsSum[myBag]+=bags[myBag][i];

	//uvecaj deljenu promenljivu
	pthread_mutex_lock(&moneyMutex);
	
	moneySum+=bagsSum[myBag];

	pthread_mutex_unlock(&moneyMutex);

	return 0;
}

int main(int argc, char *argv[]) {

	pthread_t* threads;
	pthread_attr_t attributes;
	int t, rc, i,j, n;

	//postavi eksplicitno atribut JOINABLE
	pthread_attr_init(&attributes);
	pthread_attr_setdetachstate(&attributes, PTHREAD_CREATE_JOINABLE);

	//inicijalizacija brave
	pthread_mutex_init(&moneyMutex, NULL);

	printf("Unesite broj ukupan broj dzakova za prebrojavanje: ");
	scanf("%d",&n);

	bags = malloc(n*sizeof(int*));
	bagsCount = malloc(n*sizeof(int));
	bagsSum = malloc(n*sizeof(int));
	threads = malloc(n*sizeof(pthread_t));

	printf("Unesite podatke o dzakovima:\n");
	for(i=0;i<n;i++) {
		printf("\nBroj novcanica u %d. dzaku: ",i);
		scanf("%d",bagsCount+i);
		
		bags[i] = malloc(bagsCount[i]*sizeof(int));
		
		printf("Unesite vrednost novcanica: \n");
		for(j=0;j<bagsCount[i];j++)
			scanf("%d",*(bags+i)+j);	
	}

	for(t=0;t<n;t++)
	{
		printf("Stvaram POSIX nit %d\n", t);
		// Napravi POSIX nit.
		rc = pthread_create(&threads[t], NULL, countTheMoney, (void *)t);
		if (rc)
		{
			printf("GREKA: povratna vrednost iz pthread_create(): %d\n", rc);
			exit(-1);
		}
	}

	//sacekaj da sve niti zavrse posao 
	for (i=0; i < n; i++) {
		pthread_join(threads[i], NULL);
	}
	
	//ispisi dobijene rezultate
	printf("Suma novaca po pojedinacnim dzakovima je: \n");
	for(i=0;i<n;i++)
		printf("%d ",bagsSum[i]);

	printf("\nUkupna suma novca u dzakovima je: %d",moneySum);

	//Oslobodi zauzete resurse	
	free(threads);
	free(bagsSum);
	free(bagsCount);
	for(i=0;i<n;i++)
		free(bags[i]);
	free(bags);
	
	pthread_mutex_destroy(&moneyMutex);
	pthread_attr_destroy(&attributes);

	pthread_exit(NULL);
	return 0;
}