/******************************************************************************/
/**
@file "Pthreads/C/Primer 2 - stack/stack_size.c"
@brief Primer manipulacije atributima niti

Primer koji demonstrira upotrebu PThreads API funkcija za manipulaciju
atributima niti, konkretno dohvatanje i promenu veliine steka niti.

Autor ovog primera je Blaise Barney. Kod je preuzet iz Pthreads tutorijala sa
adrese http://www.llnl.gov/computing/tutorials/pthreads/ i prilagodjen
potrebama ovog kursa.
*/
/******************************************************************************/

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

/******************************************************************************/

/**
@def N_THREADS 
@brief Broj niti koje treba kreirati
*/
#define N_THREADS	4

/**
@def N 
@brief Multiplikativna konstanta
*/
#define N		1000

/**
@def MEG_EXTRA 
@brief Aditivna konstanta
*/
#define MEG_EXTRA	1000000
 
/******************************************************************************/

/**
@var attr
@brief Struktura koja definie atribute niti
*/
pthread_attr_t attr;
 
/******************************************************************************/

/**
@brief Posao koji treba da izvri svaka od niti
@param thread_id Pokaziva na ID broj niti
@return void* Rezultat izvravanja funkcije pthread_exit

Koristan posao koji svaka od niti treba da odradi je da dohvati veliinu
svog steka i da ispie poruku na standardnom izlazu. Ostatak posla ini
popunjavanje matrice @ref N x @ref N , koja niemu ne slui.
*/
void* do_work(void* thread_id)
{
	size_t my_stack_size;
    int i, j;
    double A[N][N];

	// Dohvata velicinu steka niti.
	pthread_attr_getstacksize (&attr, &my_stack_size);
	printf("Thread %d: stack size = %d bytes \n", thread_id, my_stack_size);

	for (i=0; i<N; i++)
		for (j=0; j<N; j++)
			A[i][j] = ((i * j) / 3.452) + (N - i);

	pthread_exit(NULL);
}

/******************************************************************************/

/**
@brief Glavna nit programa
@param argc Broj argumenata komandne linije
@param argv Argumenti komandne linije
@return int Rezultat izvravanja programa

U glavnom programu se prvo pozivom funkcije pthread_attr_init inicijalizuju
atributi niti, a potom se pozivima funkcija pthread_attr_getstacksize i 
pthread_attr_setstacksize povea stek niti. Zatim se u @b for ciklusu kreiraju
niti sa novom, poveanom veliinom steka.
*/
int main(int argc, char* argv[])
{
	pthread_t threads[N_THREADS];
	size_t stack_size;
	int rc, t;
	
#ifdef WIN32
    pthread_win32_process_attach_np();
#endif

	pthread_attr_init(&attr);
	
	// Dohvatamo podrazumevanu velicinu steka niti i ispisemo je.
	pthread_attr_getstacksize (&attr, &stack_size);
	printf("Default stack size = %d\n", stack_size);
	// Izvrsimo prosirivanje steka niti.
	stack_size = sizeof(double) * N * N + MEG_EXTRA;
	printf("Amount of stack needed per thread = %d\n",stack_size);
	pthread_attr_setstacksize (&attr, stack_size);
	
	printf("Creating threads with stack size = %d bytes\n",stack_size);
	for(t = 0; t < N_THREADS; t++)
	{
		rc = pthread_create(&threads[t], &attr, do_work, (void*) t);
	
		if (rc)
		{
			printf("ERROR; return code from pthread_create() is %d\n", rc);
			exit(-1);
		}
	}

	printf("Created %d threads.\n", t);
	pthread_exit(NULL);

#ifdef WIN32
    pthread_win32_process_detach_np();
#endif
}

/******************************************************************************/
