/******************************************************************************/
/**
@file "Pthreads/C/Primer 3 - join/print_hello.c"
@brief Demonstracija upotrebe Pthread rutine join

"Hello World" program koji korienjem Pthread API-ja demonstrira kreiranje
i unitavanje niti sa ekanjem na kraj njenog izvravanja.@n @n

Autor ovog primera je <b>Blaise Barney</b>. Kod je preuzet iz Pthreads tutorijala sa
adrese http://www.llnl.gov/computing/tutorials/pthreads/ i prilagodjen
potrebama ovog kursa.
*/
/******************************************************************************/

#include <pthread.h>
#include <stdio.h>

/******************************************************************************/

/**
@def NUM_THREADS 
@brief Broj niti koje treba kreirati.
*/
#define NUM_THREADS	5

/******************************************************************************/

/**
@brief Posao koji svaka od niti treba da odradi.
@param thread_id Generiki pokaziva na redni broj niti.
@return void* Rezultat izvravanja rutine ptrhead_exit.

Na standardnom izlazu ispisuje pozdravnu poruku, svoj ID broj i svoj redni
broj.
*/
void* print_hello(void* thread_id)
{
	printf("\n%x:%x: Hello World!\n", pthread_self(), thread_id);
	pthread_exit(NULL);
}

/******************************************************************************/

/**
@brief Glavna nit programa.
@param argc Broj argumenata komandne linije.
@param argv Argumenti komandne linije.
@return int Rezultat izvravanja programa.

Kreira se @ref NUM_THREADS niti i eka se na kraj njihovog izvravanja.
*/
int main(int argc, char* argv[])
{
	pthread_t threads[NUM_THREADS];
	int t, status;

#ifdef WIN32
    pthread_win32_process_attach_np();
#endif

	for (t = 0; t < NUM_THREADS; t++)
	{
		printf("Creating thread %d\n", t);
		pthread_create(&threads[t], NULL, print_hello, (void*) t);
	}
	
	for (t = 0; t < NUM_THREADS; t++)
	{
		printf("Joining thread %d\n", t);
		pthread_join(threads[t], (void**) &status);
	}

#ifdef WIN32
    pthread_win32_process_detach_np();
#endif
}

/******************************************************************************/
