/******************************************************************************/
/**
@file "Pthreads/C/Primer 4 - mutex/dot_product_simple.c"
@brief Sekvencijalna verzija.

Jednostavan sekvencijalni program koji izraunava skalarni proizvod dva
vektora. Verzija sa nitima moe se nai u fajlu dot_product_threaded_mutex.c @n @n

Autor ovog primera je Vijay Sonnad, IBM. Najnovija verzija koda, koju je kreirao
Blaise Barney, je preuzeta iz Pthreads tutorijala sa adrese
http://www.llnl.gov/computing/tutorials/pthreads/ i prilagodjen potrebama
ovog kursa.
*/
/******************************************************************************/

#include <stdio.h>
#include <stdlib.h>

/******************************************************************************/

/**
@struct DOT_DATA
@brief Neke neophodne informacije potrebne dot_prod funkciji.

Ova struktura sadri neophodne informacije koje dozvoljavaju funkciji
dot_prod da pristupi njenim ulaznim podacima i da sauva rezultate kako bi
im se moglo kasnije pristupiti.
*/
typedef struct {
    /**
    @property a
    @brief Pokaziva na prvi cinilac.
    */
	double* a;

    /**
    @property b
    @brief Pokaziva na drugi cinilac.
    */
	double* b;

    /**
    @property sum
    @brief Suma.
    */
	double sum; 

    /**
    @property vec_len
    @brief Duina inilaca.
    */
	int vec_len; 
} DOT_DATA;

/******************************************************************************/

/**
@def VEC_LEN
@brief Duina vektora.
*/
#define VEC_LEN 100

/******************************************************************************/

/**
@var dot_str
@brief Promenljiva koja sadri podatke za obradu i rezultat obrade
*/
DOT_DATA dot_str; 

/******************************************************************************/

/**
@brief Funkcija za izraunavanje skalarnog proizvoda.

Ovu funkciju koristimo za raunanje skalarnog proizvoda.
Svi ulazni podaci bie dohvaeni pomou strukture tipa DOT_DATA
i svi dobijeni rezultati bie vraeni u istu strukturu.
Ovaj pristup nije od neke preterane koristi u sekvencijalnom 
programu, ali e se pokazati kao veoma koristan u modifikaciji
u kojoj se izraunavanje vri paralelno.
*/
void dot_prod(void)
{
	int start, end, i; 
	double my_sum, *x, *y;
	
	start = 0;
	end = dot_str.vec_len;
	x = dot_str.a;
	y = dot_str.b;

	my_sum = 0;

	for (i = start; i < end ; i++) 
	{
		my_sum += (x[i] * y[i]);
	}

	dot_str.sum = my_sum;
	
	return ;
}

/******************************************************************************/

/**
@brief Glavni program.
@param argc Broj argumenata komandne linije.
@param argv Argumenti komandne linije.
@return Vraa rezultat izvravanja programa.

Glavni program inicijalizuje promenljive i poziva dot_prod funkciju.
Na kraju, ispisuje rezultat na standardnom izlazu.
*/
int main (int argc, char* argv[])
{
	int i, len;
	double* a, *b;

	len = VEC_LEN;
	a = (double*) malloc(len * sizeof(double));
	b = (double*) malloc(len * sizeof(double));
	
	for (i = 0; i < len; i++)
	{
		a[i] = 1;
		b[i] = a[i];
	}
	
	dot_str.vec_len = len; 
	dot_str.a = a; 
	dot_str.b = b; 
	dot_str.sum = 0;
	
	dot_prod ();
	
	printf ("Sum =  %f \n", dot_str.sum);
	free (a);
	free (b);
}

/******************************************************************************/
